import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import 'package:cryptography/cryptography.dart';
import 'package:cryptography_flutter/cryptography_flutter.dart';
import 'package:flutter/foundation.dart';

class NUICryptographyBloc{

  static Uint8List randomBytes(int length) {
    Uint8List buffer = Uint8List(length);
    Random range = Random.secure();

    for (int i = 0; i < length; i++) {
      buffer[i] = range.nextInt(256);
    }

    return buffer;
  }

  Future<SecretKey> _deriveKey(String rawKey, int rawKeyLength, List<int> salt, int iteration) async {
    final pbkdf2 = Pbkdf2(
      macAlgorithm: Hmac.sha256(),
      iterations: iteration,
      bits: rawKeyLength * 8,
    );

    SecretKey secret = await pbkdf2.deriveKey(
      secretKey: SecretKey(utf8.encode(rawKey)),
      nonce: salt,
    );

    return secret;
  }

  Future<List<int>> aes256EncryptPlain({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required String data,
    required int iteration,
    required int saltLength,
    required int ivLength
  }) async {
    final salt = randomBytes(saltLength);
    final iv = randomBytes(ivLength);
    final secretKey = await _deriveKey(rawKey, rawKeyLength, salt, iteration);
    final secretBox = await algorithm.encrypt(
      utf8.encode(data),
      secretKey: secretKey,
      nonce: iv,
    );
    final List<int> result = salt + secretBox.nonce + secretBox.cipherText + secretBox.mac.bytes;
    return result;
  }

  Future<List<int>> aes256DecryptPlain({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required String data,
    required int iteration,
    required int saltLength,
    required int ivLength
  }) async {
    final cText = base64.decode(data);
    final salt = cText.sublist(0, saltLength);
    final iv = cText.sublist(saltLength, saltLength + ivLength);
    final mac = cText.sublist(cText.length - saltLength);
    final text = cText.sublist(saltLength + ivLength, cText.length - saltLength);
    final key = await _deriveKey(rawKey, rawKeyLength, salt, iteration);

    final secretBox = SecretBox(text, nonce: iv, mac: Mac(mac));
    final cleartext = await algorithm.decrypt(
      secretBox,
      secretKey: key,
    );
    return cleartext;
  }

  Future<List<int>> aes256EncryptBytes({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required Uint8List data,
    required int iteration,
    required int saltLength,
    required int ivLength
  }) async {
    final salt = randomBytes(saltLength);
    final iv = randomBytes(ivLength);
    final secretKey = await _deriveKey(rawKey, rawKeyLength, salt, iteration);
    final secretBox = await algorithm.encrypt(
      data,
      secretKey: secretKey,
      nonce: iv,
    );
    final List<int> result = salt + secretBox.nonce + secretBox.cipherText + secretBox.mac.bytes;
    return result;
  }

  Future<List<int>> aes256DecryptBytes({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required Uint8List data,
    required int iteration,
    required int saltLength,
    required int ivLength
  }) async {
    final cText = data;
    final salt = cText.sublist(0, saltLength);
    final iv = cText.sublist(saltLength, saltLength + ivLength);
    final mac = cText.sublist(cText.length - saltLength);
    final text = cText.sublist(saltLength + ivLength, cText.length - saltLength);
    final key = await _deriveKey(rawKey, rawKeyLength, salt, iteration);

    final secretBox = SecretBox(text, nonce: iv, mac: Mac(mac));
    final cleartext = await algorithm.decrypt(
      secretBox,
      secretKey: key,
    );
    return cleartext;
  }

  Future<Stream<List<int>>> aes256EncryptBytesAsStream({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required Stream<List<int>> data,
  }) async {

    final secretKey = await _deriveKey(rawKey, rawKeyLength, [], 5);
    final encryptedStream = await algorithm.encryptStream(
        data,
        secretKey: secretKey,
        nonce: [],
        onMac: (mac){}
    );
    return encryptedStream;
  }

  Future<Stream<List<int>>> aes256DecryptBytesAsStream({
    required FlutterAesGcm algorithm,
    required String rawKey,
    required int rawKeyLength,
    required Stream<List<int>> data,
  }) async {

    final secretKey = await _deriveKey(rawKey, rawKeyLength, [], 5);
    final encryptedStream = algorithm.decryptStream(
        data,
        secretKey: secretKey,
        nonce: [],
        mac: Mac.empty
    );

    return encryptedStream;
  }

}